@@ -18,7 +18,6 @@
 #include <asm/random.h>
 #include <asm/setup.h>
 #include <asm/shstk.h>
-#include <mach_apic.h>
 #include <public/sysctl.h> /* for XEN_INVALID_{SOCKET,CORE}_ID */
 
 #include "cpu.h"
@@ -439,13 +438,6 @@ static void generic_identify(struct cpuinfo_x86 *c)
 	c->apicid = phys_pkg_id((ebx >> 24) & 0xFF, 0);
 	c->phys_proc_id = c->apicid;
 
-	/*
-	 * Early init of Self Snoop support requires 0x1.edx, while there also
-	 * set 0x1.ecx as the value is in context.
-	 */
-	c->x86_capability[FEATURESET_1c] = ecx;
-	c->x86_capability[FEATURESET_1d] = edx;
-
 	eax = cpuid_eax(0x80000000);
 	if ((eax >> 16) == 0x8000)
 		c->extended_cpuid_level = eax;
@@ -486,9 +478,7 @@ static void generic_identify(struct cpuinfo_x86 *c)
 	if (c->extended_cpuid_level >= 0x80000008)
 		c->x86_capability[FEATURESET_e8b] = cpuid_ebx(0x80000008);
 	if (c->extended_cpuid_level >= 0x80000021)
-		cpuid(0x80000021,
-		      &c->x86_capability[FEATURESET_e21a], &tmp,
-		      &c->x86_capability[FEATURESET_e21c], &tmp);
+		c->x86_capability[FEATURESET_e21a] = cpuid_eax(0x80000021);
 
 	/* Intel-defined flags: level 0x00000007 */
 	if (c->cpuid_level >= 7) {
