@@ -16,9 +16,7 @@
 
 #include <xen/err.h>
 #include <xen/init.h>
-#include <xen/lib.h>
 #include <xen/mm.h> /* TODO: Fix asm/tlbflush.h breakage */
-#include <xen/sha2.h>
 
 #include <asm/msr.h>
 
@@ -42,10 +40,7 @@ struct microcode_patch {
     uint8_t  mc_patch_data_id[2];
     uint8_t  mc_patch_data_len;
     uint8_t  init_flag;
-    union {
-        uint32_t checksum; /* Fam12h and earlier */
-        uint32_t min_rev;  /* Zen3-5, post Entrysign */
-    };
+    uint32_t mc_patch_data_checksum;
     uint32_t nb_dev_id;
     uint32_t sb_dev_id;
     uint16_t processor_rev_id;
@@ -57,11 +52,11 @@ struct microcode_patch {
 };
 
 #define UCODE_MAGIC                0x00414d44
-#define UCODE_EQUIV_CPU_TABLE_TYPE 0x00000000
+#define UCODE_EQUIV_TYPE           0x00000000
 #define UCODE_UCODE_TYPE           0x00000001
 
 struct container_equiv_table {
-    uint32_t type; /* UCODE_EQUIV_CPU_TABLE_TYPE */
+    uint32_t type; /* UCODE_EQUIV_TYPE */
     uint32_t len;
     struct equiv_cpu_entry eq[];
 };
@@ -96,64 +91,6 @@ static struct {
     uint16_t id;
 } equiv __read_mostly;
 
-static const struct patch_digest {
-    uint32_t patch_id;
-    uint8_t digest[SHA2_256_DIGEST_SIZE];
-} patch_digests[] = {
-#include "amd-patch-digests.c"
-};
-static bool __ro_after_init entrysign_mitigiated_in_firmware;
-
-static int cf_check cmp_patch_id(const void *key, const void *elem)
-{
-    const struct patch_digest *pd = elem;
-    uint32_t patch_id = *(uint32_t *)key;
-
-    if ( patch_id == pd->patch_id )
-        return 0;
-    else if ( patch_id < pd->patch_id )
-        return -1;
-    return 1;
-}
-
-static bool check_digest(const struct container_microcode *mc)
-{
-    const struct microcode_patch *patch = mc->patch;
-    const struct patch_digest *pd;
-    uint8_t digest[SHA2_256_DIGEST_SIZE];
-
-    /*
-     * Zen1 thru Zen5 CPUs are known to use a weak signature algorithm on
-     * microcode updates.  If this has not been mitigated in firmware, check
-     * the digest of the patch against a list of known provenance.
-     */
-    if ( boot_cpu_data.x86 < 0x17 || boot_cpu_data.x86 > 0x1a ||
-         entrysign_mitigiated_in_firmware || !opt_digest_check )
-        return true;
-
-    pd = bsearch(&patch->patch_id, patch_digests, ARRAY_SIZE(patch_digests),
-                 sizeof(struct patch_digest), cmp_patch_id);
-    if ( !pd )
-    {
-        printk(XENLOG_WARNING "No digest found for patch_id %08x\n",
-               patch->patch_id);
-        return false;
-    }
-
-    sha2_256_digest(digest, patch, mc->len);
-
-    if ( memcmp(digest, pd->digest, sizeof(digest)) )
-    {
-        printk(XENLOG_WARNING "Patch %08x SHA256 mismatch:\n"
-               "  expected %" STR(SHA2_256_DIGEST_SIZE) "phN\n"
-               "       got %" STR(SHA2_256_DIGEST_SIZE) "phN\n",
-               patch->patch_id, pd->digest, digest);
-        return false;
-    }
-
-    return true;
-}
-
 static void cf_check collect_cpu_info(void)
 {
     struct cpu_signature *csig = &this_cpu(cpu_sig);
@@ -237,8 +174,7 @@ static bool check_final_patch_levels(const struct cpu_signature *sig)
     return false;
 }
 
-static enum microcode_match_result compare_revisions(
-    uint32_t old_rev, uint32_t new_rev)
+static int compare_revisions(uint32_t old_rev, uint32_t new_rev)
 {
     if ( new_rev > old_rev )
         return NEW_UCODE;
@@ -249,90 +185,52 @@ static enum microcode_match_result compare_revisions(
     return OLD_UCODE;
 }
 
-static enum microcode_match_result microcode_fits(
-    const struct microcode_patch *patch)
+/*
+ * Check whether this microcode patch is applicable for the current CPU.
+ *
+ * AMD microcode blobs only have the "equivalent CPU identifier" which is a 16
+ * bit contraction of the 32 bit Family/Model/Stepping.
+ *
+ * We expect to only be run after scan_equiv_cpu_table() has found a valid
+ * mapping for the current CPU.  If this is violated, the 0 in equiv.id will
+ * cause the patch to be rejected too.
+ */
+static bool microcode_fits_cpu(const struct microcode_patch *patch)
 {
-    unsigned int cpu = smp_processor_id();
-    const struct cpu_signature *sig = &per_cpu(cpu_sig, cpu);
+    ASSERT(equiv.sig);
 
-    if ( equiv.sig != sig->sig ||
-         equiv.id  != patch->processor_rev_id )
-        return MIS_UCODE;
-
-    return compare_revisions(sig->rev, patch->patch_id);
+    return equiv.id == patch->processor_rev_id;
 }
 
-static enum microcode_match_result compare_header(
-    const struct microcode_patch *new, const struct microcode_patch *old)
+static int cf_check amd_compare(
+    const struct microcode_patch *old, const struct microcode_patch *new)
 {
-    if ( new->processor_rev_id != old->processor_rev_id )
-        return MIS_UCODE;
+    /* Both patches to compare are supposed to be applicable to local CPU. */
+    ASSERT(microcode_fits_cpu(new));
+    ASSERT(microcode_fits_cpu(old));
 
     return compare_revisions(old->patch_id, new->patch_id);
 }
 
-/*
- * Check whether this patch has a minimum revision given, and whether the
- * condition is satisfied.
- *
- * In linux-firmware for CPUs suffering from the Entrysign vulnerability,
- * ucodes signed with the updated signature algorithm have reused the checksum
- * field as a min-revision field.  From public archives, the checksum field
- * appears to have been unused since Fam12h.
- *
- * Returns false if there is a min revision given, and it suggests that that
- * the patch cannot be loaded on the current system.  True otherwise.
- */
-static bool check_min_rev(const struct microcode_patch *patch)
+static int cf_check apply_microcode(const struct microcode_patch *patch,
+                                    unsigned int flags)
 {
-    ASSERT(microcode_fits(patch));
-
-    if ( patch->processor_rev_id < 0xa000 || /* pre Zen3? */
-         patch->min_rev == 0 )               /* No min rev specified */
-        return true;
-
-    /*
-     * Sanity check, as this is a reused field.  If this is a true
-     * min_revision field, it will differ only in the bottom byte from the
-     * patch_id.  Otherwise, it's probably a checksum.
-     */
-    if ( (patch->patch_id ^ patch->min_rev) & ~0xff )
-    {
-        printk(XENLOG_WARNING
-               "microcode: patch %#x has unexpected min_rev %#x\n",
-               patch->patch_id, patch->min_rev);
-        return true;
-    }
-
-    return this_cpu(cpu_sig).rev >= patch->min_rev;
-}
-
-static enum microcode_match_result cf_check compare_patch(
-    const struct microcode_patch *new, const struct microcode_patch *old)
-{
-    /* Both patches to compare are supposed to be applicable to local CPU. */
-    ASSERT(microcode_fits(new) != MIS_UCODE);
-    ASSERT(microcode_fits(old) != MIS_UCODE);
-
-    return compare_header(new, old);
-}
-
-static int cf_check apply_microcode(const struct microcode_patch *patch)
-{
-    int hw_err;
+    int hw_err, result;
     unsigned int cpu = smp_processor_id();
     struct cpu_signature *sig = &per_cpu(cpu_sig, cpu);
     uint32_t rev, old_rev = sig->rev;
-    enum microcode_match_result result = microcode_fits(patch);
+    bool ucode_force = flags & XENPF_UCODE_FORCE;
 
-    if ( result == MIS_UCODE )
+    if ( !microcode_fits_cpu(patch) )
         return -EINVAL;
 
+    result = compare_revisions(old_rev, patch->patch_id);
+
     /*
      * Allow application of the same revision to pick up SMT-specific changes
      * even if the revision of the other SMT thread is already up-to-date.
      */
-    if ( result == OLD_UCODE )
+    if ( !ucode_force && (result == SAME_UCODE || result == OLD_UCODE) )
         return -EEXIST;
 
     if ( check_final_patch_levels(sig) )
@@ -343,14 +241,6 @@ static int cf_check apply_microcode(const struct microcode_patch *patch)
         return -ENXIO;
     }
 
-    if ( !check_min_rev(patch) )
-    {
-        printk(XENLOG_ERR
-               "microcode: CPU%u current rev %#x below patch min_rev %#x\n",
-               cpu, sig->rev, patch->min_rev);
-        return -ENXIO;
-    }
-
     hw_err = wrmsr_safe(MSR_AMD_PATCHLOADER, (unsigned long)patch);
 
     /* get patch id after patching */
@@ -376,6 +266,8 @@ static int cf_check apply_microcode(const struct microcode_patch *patch)
            "microcode: CPU%u updated from revision %#x to %#x, date = %04x-%02x-%02x\n",
            cpu, old_rev, rev, patch->year, patch->month, patch->day);
 
+    amd_check_zenbleed();
+
     return 0;
 }
 
@@ -442,10 +334,10 @@ static struct microcode_patch *cf_check cpu_request_microcode(
         buf  += 4;
         size -= 4;
 
-        if ( size < sizeof(*et) ||
-             (et = buf)->type != UCODE_EQUIV_CPU_TABLE_TYPE ||
-             size - sizeof(*et) < et->len ||
-             et->len % sizeof(et->eq[0]) )
+        if ( size < sizeof(*et) ||                   /* No space for header? */
+             (et = buf)->type != UCODE_EQUIV_TYPE || /* Not an Equivalence Table? */
+             size - sizeof(*et) < et->len ||         /* No space for table? */
+             et->len % sizeof(et->eq[0]) )           /* Not multiple of equiv_cpu_entry? */
         {
             printk(XENLOG_ERR "microcode: Bad equivalent cpu table\n");
             error = -EINVAL;
@@ -458,7 +350,12 @@ static struct microcode_patch *cf_check cpu_request_microcode(
 
         error = scan_equiv_cpu_table(et);
 
-        /* -ESRCH means no applicable microcode in this container. */
+        /*
+         * -ESRCH means no applicable microcode in this container.  But, there
+         * might be subsequent containers in the blob.  Skipping to the end of
+         * this container still requires us to follow the UCODE_UCODE_TYPE/len
+         * metadata because there's no overall container length given.
+         */
         if ( error && error != -ESRCH )
             break;
         skip_ucode = error;
@@ -468,10 +365,10 @@ static struct microcode_patch *cf_check cpu_request_microcode(
         {
             const struct container_microcode *mc;
 
-            if ( size < sizeof(*mc) ||
-                 (mc = buf)->type != UCODE_UCODE_TYPE ||
-                 size - sizeof(*mc) < mc->len ||
-                 mc->len < sizeof(struct microcode_patch) )
+            if ( size < sizeof(*mc) ||                      /* No space for container header? */
+                 (mc = buf)->type != UCODE_UCODE_TYPE ||    /* Not a ucode blob? */
+                 size - sizeof(*mc) < mc->len ||            /* No space for blob? */
+                 mc->len < sizeof(struct microcode_patch) ) /* No space for patch header? */
             {
                 printk(XENLOG_ERR "microcode: Bad microcode data\n");
                 error = -EINVAL;
@@ -498,9 +395,10 @@ static struct microcode_patch *cf_check cpu_request_microcode(
              * If the new ucode covers current CPU, compare ucodes and store the
              * one with higher revision.
              */
-            if ( (microcode_fits(mc->patch) != MIS_UCODE) &&
-                 (!saved || (compare_header(mc->patch, saved) == NEW_UCODE)) &&
-                 check_digest(mc) )
+            if ( microcode_fits_cpu(mc->patch) &&
+                 (!saved ||
+                  compare_revisions(saved->patch_id,
+                                    mc->patch->patch_id) == NEW_UCODE) )
             {
                 saved = mc->patch;
                 saved_size = mc->len;
@@ -546,126 +444,21 @@ static struct microcode_patch *cf_check cpu_request_microcode(
     return patch;
 }
 
+static const char __initconst amd_cpio_path[] =
+    "kernel/x86/microcode/AuthenticAMD.bin";
+
 static const struct microcode_ops __initconst_cf_clobber amd_ucode_ops = {
     .cpu_request_microcode            = cpu_request_microcode,
     .collect_cpu_info                 = collect_cpu_info,
     .apply_microcode                  = apply_microcode,
-    .compare_patch                    = compare_patch,
+    .compare                          = amd_compare,
+    .cpio_path                        = amd_cpio_path,
 };
 
 void __init ucode_probe_amd(struct microcode_ops *ops)
 {
-    /*
-     * The Entrysign vulnerability (SB-7033, CVE-2024-36347) affects Zen1-5
-     * CPUs.  Taint Xen if digest checking is turned off.
-     */
-    if ( boot_cpu_data.x86 >= 0x17 && boot_cpu_data.x86 <= 0x1a &&
-         !opt_digest_check )
-    {
-        printk(XENLOG_WARNING
-               "Microcode patch additional digest checks disabled\n");
-        add_taint(TAINT_CPU_OUT_OF_SPEC);
-    }
-
     if ( boot_cpu_data.x86 < 0x10 )
         return;
 
     *ops = amd_ucode_ops;
 }
-
-#if 0 /* Manual CONFIG_SELF_TESTS */
-static void __init __constructor test_digests_sorted(void)
-{
-    for ( unsigned int i = 1; i < ARRAY_SIZE(patch_digests); ++i )
-    {
-        if ( patch_digests[i - 1].patch_id < patch_digests[i].patch_id )
-            continue;
-
-        panic("patch_digests[] not sorted: %08x >= %08x\n",
-              patch_digests[i - 1].patch_id,
-              patch_digests[i].patch_id);
-    }
-}
-#endif /* CONFIG_SELF_TESTS */
-
-/*
- * The Entrysign vulnerability affects all Zen1 thru Zen5 CPUs.  Firmware
- * fixes were produced from Nov 2024.  Zen3 thru Zen5 can continue to take
- * OS-loadable microcode updates using a new signature scheme, as long as
- * firmware has been updated first.
- */
-void __init amd_check_entrysign(void)
-{
-    unsigned int curr_rev;
-    uint8_t fixed_rev;
-
-    if ( boot_cpu_data.x86_vendor != X86_VENDOR_AMD ||
-         boot_cpu_data.x86 < 0x17 ||
-         boot_cpu_data.x86 > 0x1a )
-        return;
-
-    /*
-     * Table taken from Linux, which is the only known source of information
-     * about client revisions.  Note, Linux expresses "last-vulnerable-rev"
-     * while Xen wants "first-fixed-rev".
-     */
-    curr_rev = this_cpu(cpu_sig).rev;
-    switch ( curr_rev >> 8 )
-    {
-    case 0x080012: fixed_rev = 0x78; break;
-    case 0x080082: fixed_rev = 0x10; break;
-    case 0x083010: fixed_rev = 0x7d; break;
-    case 0x086001: fixed_rev = 0x0f; break;
-    case 0x086081: fixed_rev = 0x09; break;
-    case 0x087010: fixed_rev = 0x35; break;
-    case 0x08a000: fixed_rev = 0x0b; break;
-    case 0x0a0010: fixed_rev = 0x7b; break;
-    case 0x0a0011: fixed_rev = 0xdb; break;
-    case 0x0a0012: fixed_rev = 0x44; break;
-    case 0x0a0082: fixed_rev = 0x0f; break;
-    case 0x0a1011: fixed_rev = 0x54; break;
-    case 0x0a1012: fixed_rev = 0x4f; break;
-    case 0x0a1081: fixed_rev = 0x0a; break;
-    case 0x0a2010: fixed_rev = 0x30; break;
-    case 0x0a2012: fixed_rev = 0x13; break;
-    case 0x0a4041: fixed_rev = 0x0a; break;
-    case 0x0a5000: fixed_rev = 0x14; break;
-    case 0x0a6012: fixed_rev = 0x0b; break;
-    case 0x0a7041: fixed_rev = 0x0a; break;
-    case 0x0a7052: fixed_rev = 0x09; break;
-    case 0x0a7080: fixed_rev = 0x0a; break;
-    case 0x0a70c0: fixed_rev = 0x0a; break;
-    case 0x0aa001: fixed_rev = 0x17; break;
-    case 0x0aa002: fixed_rev = 0x19; break;
-    case 0x0b0021: fixed_rev = 0x47; break;
-    case 0x0b0081: fixed_rev = 0x12; break;
-    case 0x0b1010: fixed_rev = 0x47; break;
-    case 0x0b2040: fixed_rev = 0x32; break;
-    case 0x0b4040: fixed_rev = 0x32; break;
-    case 0x0b4041: fixed_rev = 0x02; break;
-    case 0x0b6000: fixed_rev = 0x32; break;
-    case 0x0b6080: fixed_rev = 0x32; break;
-    case 0x0b7000: fixed_rev = 0x32; break;
-    default:
-        printk(XENLOG_WARNING
-               "Unrecognised CPU %02x-%02x-%02x ucode 0x%08x, assuming vulnerable to Entrysign\n",
-               boot_cpu_data.x86, boot_cpu_data.x86_model,
-               boot_cpu_data.x86_mask, curr_rev);
-        return;
-    }
-
-    /*
-     * This check is best-effort.  If the platform looks to be out of date, it
-     * probably is.  If the platform looks to be fixed, it either genuinely
-     * is, or malware has gotten in before Xen booted and all bets are off.
-     */
-    if ( (uint8_t)curr_rev >= fixed_rev )
-    {
-        entrysign_mitigiated_in_firmware = true;
-        return;
-    }
-
-    printk(XENLOG_WARNING
-           "WARNING: Platform vulnerable to Entrysign (SB-7033, CVE-2024-36347) - firmware update required\n");
-    add_taint(TAINT_CPU_OUT_OF_SPEC);
-}
