@@ -1,6 +1,8 @@
 #ifndef _EARLYCPIO_H
 #define _EARLYCPIO_H
 
+#include <xen/types.h>
+
 #define MAX_CPIO_FILE_NAME 18
 
 struct cpio_data {
@@ -9,7 +11,6 @@ struct cpio_data {
 	char name[MAX_CPIO_FILE_NAME];
 };
 
-struct cpio_data find_cpio_data(const char *path, void *data, size_t len,
-				long *offset);
+struct cpio_data find_cpio_data(const char *path, void *data, size_t len);
 
 #endif /* _EARLYCPIO_H */
